/* -*- Mode: C++; c-basic-offset: 4; indent-tabs-mode: nil; tab-width: 4 -*- */
/* vi: set ts=4 sw=4 expandtab: (add to ~/.vimrc: set modeline modelines=5) */
/* This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/. */

#ifndef __nanojit_CodeAlloc__
#define __nanojit_CodeAlloc__

namespace nanojit
{

	class CodeAllocatorImpl
	{
	public:
		CodeAllocatorImpl() {}
		virtual ~CodeAllocatorImpl() {}
		//
		// CodeAlloc's SPI (Service Provider Interface). Implementations must be
		// defined by nanojit embedder. Allocation  failures should cause an exception
		// or longjmp; nanojit intentionally does not check for null.
		//

		/** allocate nbytes of memory to hold code. Never return null!  */
		virtual void* allocCodeChunk() = 0;

		/** free a block previously allocated by allocCodeMem. nbytes will
		 * match the previous allocCodeMem, but is provided here as well
		 * to mirror the mmap()/munmap() api. markCodeChunkWrite() will have
		 * been called if necessary, so it is not necessary for freeCodeChunk()
		 * to do it again. */
		virtual void freeCodeChunk(void* addr) = 0;

		/** make this specific extent ready to execute (might remove write) */
		virtual void markCodeChunkExec(void* addr) = 0;

		/** make this extent ready to modify (might remove exec) */
		virtual void markCodeChunkWrite(void* addr) = 0;

		/**
		 * Check the permission bits of the region specified by 'addr' for 'nbytes' bytes
		 * to ensure that they are set as expected for the given state. (Exec or Write)
		 * This callback is only issued in debug builds for validation purposes.
		 * It is up to the service provider to decide how to interprete the request; i.e
		 * it will most likely return true if the markCodeChunkXXX() methods are
		 * stubbed out.  Otherwise, the chunk's memory should be interrogated for
		 * correct permission bits.
		 */
		virtual bool checkChunkMark(void* addr, bool isExec) = 0;

		/** flush the icache for a specific extent */
		virtual void flushICache(void *start, size_t len) = 0;

		virtual size_t getChunkSizeInBytes() const = 0;

		virtual ptrdiff_t absoluteOffset() const = 0;
	};

	/**
	 * Code memory allocator is a long lived manager for many code blocks that
	 * manages interaction with an underlying code memory allocator,
	 * sets page permissions.  CodeAlloc provides APIs for allocating and freeing
	 * individual blocks of code memory (for methods, stubs, or compiled
	 * traces), static functions for managing lists of allocated code, and has
	 * a few pure virtual methods that embedders must implement to provide
	 * memory to the allocator.
	 *
	 * A "chunk" is a region of memory obtained from allocCodeChunk; it must
	 * be page aligned and be a multiple of the system page size.
	 *
	 * A "block" is a region of memory within a chunk.  It can be arbitrarily
	 * sized and aligned, but is always contained within a single chunk.
	 * class CodeList represents one block; the members of CodeList track the
	 * extent of the block and support creating lists of blocks.
	 *
	 * The allocator coalesces free blocks when it can, in free(), but never
	 * coalesces chunks.
	 */
	class CodeAlloc
	{
		CodeAllocatorImpl* _impl;

		/** allocated heap blocks */
		HeapBlock* heapBlockList;

		/** free blocks. */
		CodeBlockList availableBlockList;

		size_t totalAllocated;

		/** Number of bytes to request from VMPI layer, always a multiple of the page size */
		size_t bytesPerAlloc;

		bool _check_page_flags;

		/** add raw memory to the free list */
		void addMem();

	public:
		static const size_t minAllocSize = 128;


		CodeAlloc(CodeAllocatorImpl* impl, bool check_page_flags);
		~CodeAlloc();

		/** return all the memory allocated through this allocator to the gcheap. */
		void reset();

		/** return true if any allocations are still active */
		bool hasActiveAllocations() const;

		/** allocate some memory for code returning pointers to the region
		    terminator will be switched to writable state */
		CodeBlock* alloc();

		/** allocate a fixed size block, memory is set to writable
		    terminator will be switched to writable state */
		CodeBlock* allocFixedSize(size_t a_iSize);


		/** free several blocks */
		void freeList(CodeBlockList* blockList);

		/** update block instance to give back unused memory [start:codeEnd[ to the allocator		
		    terminator state must be writable and is not affected */
		CodeBlock* addRemainder(CodeBlock* block, NIns* codeEnd);

		/** flush the icache for the one block, before executing */
		void flushBlockIcache(CodeBlock* codeBlock);

		/** flush the icache for all code in the list, before executing */
		void flushListIcache(CodeBlockList* blockList);


		/** return the number of bytes in all the code blocks in "code", including block overhead */
#ifdef PERFM
		static size_t size(const CodeList* code);
#endif

		/** return the total number of bytes held by this CodeAlloc. */
		size_t size();

		/** get stats about heap usage */
		void getStats(size_t& total, size_t& frag_size, size_t& free_size);

		/** print out stats about heap usage */
		void logStats();

		/** protect all code managed by this CodeAlloc */
		void markAllExec();

		/** unprotect all code managed by this CodeAlloc */
		void markAllWrite();

		/** protect all mem in the block list */
		void markListExec(CodeBlockList* blockList);

		/** allow all write in the block list */
		void markListWrite(CodeBlockList* blockList);

#ifdef NANOJIT_DEBUG
		/** test if all terminator blocks are mark as executable */
		bool testIsAllExec();

		/** make sure all the higher/lower pointers are correct for every block */
		void sanityCheck();
#endif

		size_t getChunkSizeInBytes() const;

		/** protect an entire chunk */
		void markHeapBlockExec(HeapBlock* term);

		/** unprotect the code chunk containing just this one block */
		void markHeapBlockWrite(HeapBlock* term);

		ptrdiff_t absoluteOffset() const;

	private:
		/** free a block of memory previously returned by alloc().
		    terminator will be in executable state after this call */
		void free(CodeBlock* block);
	};

	inline size_t CodeAlloc::getChunkSizeInBytes() const {
		return _impl->getChunkSizeInBytes();
	}

	inline ptrdiff_t CodeAlloc::absoluteOffset() const {
		return _impl->absoluteOffset();
	}

}

#endif // __nanojit_CodeAlloc__
